import ij.*;
import ij.io.*;
import ij.process.*;
import ij.gui.*;
import ij.plugin.*;
import ij.plugin.filter.*;
import ij.plugin.frame.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.lang.*;
import java.io.*;
import imagescience.feature.Edges;
import imagescience.image.FloatImage;
import imagescience.image.Image;
import imagescience.image.Coordinates;
import imagescience.image.Axes;



public class CellophaneManualNoDext_ implements PlugInFilter, MouseListener, KeyListener{


    public class Snake{
	
	final int MAX_POINTS = 300;
	final int STROKE_WIDTH = 3;	
	int px[];
	int py[];
	int size;
	FloatImage snake_img;
	ImagePlus imp;
	int neb_size;
	int hneb;

	float prune_val;

	double Eint[][];
	double Eext[][];
	double Etot[][];


	double Eint_weight;

	boolean known_center;
	boolean known_extr;
	boolean known_mid;
	double center[];
	int extr[];
	int mid[];

	Random rnd;
	ArrayList<Integer> ptlist;//used for shuffling point order

	public Snake(int h){
	    size = 0;
	    px = new int[MAX_POINTS];
	    py = new int[MAX_POINTS];
	    hneb = h;
	    neb_size=2*h+1;
	
	    Eint = new double[neb_size][neb_size];
	    Eext = new double[neb_size][neb_size];
	    Etot = new double[neb_size][neb_size];
	    Eint_weight =1;
	    ptlist = new ArrayList<Integer>();
	    if(rnd ==null){
		rnd = new Random();
	    }
	    center = new double[2];
	    known_center =false;

	    extr = new int[2];
	    known_extr=false;

	    mid = new int[2];
	    known_mid=false;
	    
	    prune_val=2;
	}
	public Snake(){this(5);}

	public void empty(){size=0;known_center =known_extr=false;}

	public void setInternalEnergyWeight(double w){
	    Eint_weight = w;
	}

	public void updateInternalEnergy(int i){
	    int next = (i+1)%size;
	    int prev = i>0?i-1:size-1;
	    double middlex =  0.5*(px[next]+px[prev]);
	    double middley =  0.5*(py[next]+py[prev]);
	    double diff2_x =  0.5*(px[next]-px[prev]);
	    double diff2_y =   0.5*(py[next]-py[prev]);	
	    double diff2n =  Math.sqrt(diff2_x*diff2_x+diff2_y*diff2_y);
	    for(int j=0;j<neb_size;j++){
		int candy = py[i]-hneb+j;
		for(int k=0;k<neb_size;k++){
		    double candx = px[i]-hneb+k;
		    double tomidx = middlex -candx;
		    double tomidy = middley -candy;
		    double tomno =  Math.sqrt(tomidx*tomidx+tomidy*tomidy);
		    Eint[j][k] =2*tomno/(diff2n+tomno);
		}
	    }
	}

	public void updateExternalEnergy(int i,FloatImage fimg){
	    int width = imp.getDimensions()[0];
	    int height = imp.getDimensions()[1];
	    int nx=0,ny=0,sx=neb_size,sy=neb_size;
	    if(px[i]>=hneb && py[i]>=hneb && 
	       width-px[i]>hneb && height-py[i]>hneb){
		// nx = hneb -px[i];
		// ny = hneb -py[i];


	    }
	    else{ //neighbourhood somewhat outside of image
		 nx=px[i]>=hneb?0:hneb-px[i];
		 ny=py[i]>=hneb?0:hneb-py[i];
		 sx=width-px[i]>hneb? neb_size-nx : hneb+width-px[i];
		 sy=height-py[i]>hneb? neb_size-ny : hneb+height-py[i];
		for(int j=0;j<neb_size;j++){ //reset to 0 to avoid previous values
		    for(int k=0;k<neb_size;k++){
			Eext[j][k] = 0;
		    }
		}
	    }		
	    Coordinates co = new Coordinates(px[i]-hneb+nx,py[i]-hneb+ny);
	    fimg.get(co,Eext);
	    double mmax =0;
	     for(int j=0;j<neb_size;j++){
		for(int k=0;k<neb_size;k++){
		    if(mmax<Eext[k][j]){
			mmax = Eext[k][j];
		    }
		}
	     }
	     for(int j=0;j<neb_size;j++){
		 for(int k=0;k<neb_size;k++){
		     Eext[k][j] = 1-(Eext[k][j]/mmax);
		 }
	     }
	}


	void printEnergies(){
	    IJ.log("internal ");
	    for(int i=0;i<neb_size;i++){
		String row = " ";
		for(int j=0;j<neb_size;j++){
		    row = row+Eint[i][j]+ " ";
		}
		IJ.log(row);
	    }	
	}

	public int addPoint(int x,int y){
	    if(size<MAX_POINTS){
		//		IJ.log("adding point");
		px[size]=x;
		py[size]=y;
		ptlist.add(new Integer(size++));
		
		return 1;
	    }
	    else{
		return 0;
	    }
	}
	public int addPoint(int x,int y,int pos){
	    if(size<MAX_POINTS){
		//		IJ.log("adding point");
		ptlist.add(new Integer(size++));
		for(int j=size-1;j>pos;j--){
		    px[j]=px[j-1];
		    py[j]=py[j-1];
		}
		px[pos]=x;
		py[pos]=y;
		if(known_extr){
		    for(int a=0;a<2;a++){
			if(extr[a]>=pos){
			    extr[a]=extr[a]+1;
			}
		    }
		    if(extr[1]==size){
			extr[1] = extr[0];
			extr[0] = 0;
		    }
		}
		if(known_mid){
		    for(int a=0;a<2;a++){
			if(mid[a]>=pos){
			    mid[a]=mid[a]+1;
			}
		    }
		    if(mid[1]==size){
			mid[1] = mid[0];
			mid[0] = 0;
		    }
		}
		return 1;
	    }
	    else{
		return 0;
	    }
	}

	public int removePoint(int i){
	    if(i>=0 && i<size){
		for(int j=i+1;j<size;j++){
		    px[j-1]=px[j];
		    py[j-1]=py[j];
		}
		size -= 1;
		if(known_extr){
		    for(int a=0;a<2;a++){
			if(extr[a]>=i){
			    extr[a]=extr[a]-1;
			}
		    }
		    if(extr[0]==-1){
			extr[0] = extr[1];
			extr[1] = size-1;
		    }
		}
		if(known_mid){
		    for(int a=0;a<2;a++){
			if(mid[a]>=i){
			    mid[a]=mid[a]-1;
			}
		    }
		    if(mid[0]==-1){
			mid[0] = mid[1];
			mid[1] = size-1;
		    }
		}
		IJ.log("removing point "+i);
		return 1;
	    }
	    else{
		return 0;
	    }
	}

	public int prune(double space){
	    double space2 = space*space;
	    int nb= 0;
	    for(int i=0;i<size-1;i++){
		while(distance2(i,i+1)<=space2 && i<size-1){
		    nb+=removePoint(i+1);
		}
	    }
	    return nb;
	}
	
	void setFloatImage(FloatImage img){
	    snake_img = img;
	    snake_img.axes(3); //Axes.YX
	}

	void setImage(ImagePlus imp){
	    this.imp = imp;
	}
	
	void listPoints(){
		IJ.log("list of points");
	    for (int i=0;i<size;i++){
		IJ.log(px[i]+ " " + py[i]);
	    }
	}

	void getPointsFromRoi(Roi r, int maxpts){
	    Polygon pol = r.getPolygon();
	    float step;
	    if(maxpts>0){
		maxpts = Math.max(maxpts,MAX_POINTS-1);
		step = pol.npoints/(float)maxpts;
	    }
	    else{
		step=(float)1.0;
	    }
	    int i=0;
	    float f =0;
	    //	    snake.listPoints();
	    empty();
	    addPoint(pol.xpoints[0],pol.ypoints[0]);
	    while(f<pol.npoints){
		if((int)f > i){
		    i = (int)(f);
		    snake.addPoint(pol.xpoints[i],pol.ypoints[i]);
		}
		f += step;
	    }
	    addPoint(pol.xpoints[0],pol.ypoints[0]);
	    known_center= known_extr=false;
	}

	void updatePoint(int i,FloatImage img){
	    double fac = 2*Eint_weight*rnd.nextDouble();
	    updateInternalEnergy(i);
	    updateExternalEnergy(i,img);
	    double mmin =Double.MAX_VALUE;
	    int mink=hneb;

	    int minj=hneb;
	    for(int j=0;j<neb_size;j++){
		 for(int k=0;k<neb_size;k++){
		     Etot[j][k] = Eext[j][k]+fac*Eint[j][k];
		     if(mmin > Etot[j][k]){ // current max
			 mmin = Etot[j][k];
			 mink = k;
			 minj= j;
		     }
		 }
	    }
	    px[i] += mink-hneb;
	    py[i] += minj-hneb;
	}

	void updateSnake(FloatImage img){
	    Collections.shuffle(ptlist);
	    Iterator it = ptlist.iterator();
	    while(it.hasNext()){
		int i= ((Integer) it.next()).intValue();
		updatePoint(i,img);
	    }
	}


	PolygonRoi asRoi(){
	    PolygonRoi roi = new PolygonRoi(px,py,size,Roi.POLYLINE);
	    roi.setStrokeWidth(STROKE_WIDTH);
	    return roi;
	}

	PolygonRoi asRoi(int index, boolean dir){
	    int[] npx = new int[size+1];
	    int[] npy = new int[size+1];
	    int incr=dir?1:-1;
	    int j=index;
	    for(int i=0;i<=size;i++){
		npx[i]=px[j];
		npy[i]=py[j];
		int k = j+incr;
		j =k<0?size-(Math.abs(k)%size):k%size; //real modulo
		//		if(j<0){j=size-1;}
		//		System.out.print(" "+j);
	    }
	    PolygonRoi roi = new PolygonRoi(npx,npy,size+1,Roi.POLYLINE);
	    roi.setStrokeWidth(STROKE_WIDTH);
	    return roi;
	}


	void run(int n){
	    for(int i=0;i<n;i++){
		updateSnake(snake_img);
	    }
	    known_center =false;
	    known_extr= false;
	}

	int[] getX(){return px;}
	int[] getY(){return py;}
	int getX(int i){return px[i];}
	int getY(int i){return py[i];}

	double[] getExtBackground(int pix,ImagePlus im, boolean showback){
  //	    int pixval[4];
	    int[] bpx; //to visualize the background
	    int[] bpy; //to visualize the background
	    bpx = px.clone();
	    bpy  = py.clone();
	    updateCenter();
	    double xc=center[0];
	    double yc=center[1];	   
	  
	    for(int i=0;i<size;i++){
		double xdir = xc-px[i];
		double ydir = yc-py[i];
		double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		xdir = xdir/norm;
		ydir = ydir/norm;
		int nx = px[i]-(int)(pix*xdir);
		int ny = py[i]-(int)(pix*ydir);
		bpx[i] = nx;
		bpy[i] = ny;
	    }
	    PolygonRoi bgroi = new PolygonRoi(bpx,bpy,size,Roi.POLYLINE);
	    bgroi.setStrokeWidth(3);
	    im.setRoi(bgroi);
	    ProfilePlot bgprof = new ProfilePlot(im,true);
	    double[] bgprofval = bgprof.getProfile();
	    double backr = 0;
	    double backr2=0;  
	    for(int i=0;i<bgprofval.length;i++){
		backr += bgprofval[i];
		backr2 += bgprofval[i]*bgprofval[i];
	    }
	    backr = backr/bgprofval.length; //mean value
	    backr2 = backr2/bgprofval.length;//mean squared value
	    backr2 = Math.sqrt(backr2 - backr*backr);// to get the s.d.
	    if(showback){
		bgroi.setName("background");
		rois.addRoi(bgroi);
	    }
	    //	    IJ.log("background: "+ backr);
	    double[] ret = {backr,backr2};
	    return ret;
	}


	int findClosestPoint(float x1,float y1, float x2, float y2, boolean add){
	    int mini = findClosestPoint(x1,y1,false);
	    float b1 = x1-x2;
	    float b2 = y1-y2;
	    float Ai1=b2,Ai2=-b1;
	    boolean done = false;
	    while(done==false){
		    int k=mini+1; 		    
		    int l =k<0?size-(Math.abs(k)%size):k%size; //real modulo
		    float Ai4 = -px[l]+px[mini];
		    float Ai3 = py[l]-py[mini];
		    float det = Ai1*Ai4-Ai2*Ai3;

		    float p1 = px[mini]-x2;
		    float p2 = py[mini]-y2;

		    float k1 = (p1*Ai1+p2*Ai2)/det;
		    float k2 = (p1*Ai3+p2*Ai3)/det;
		    if(k1<0){
			mini--;
			if(mini<0){mini = size-1;}
		    }
		    else{
			if(k1>1){
			    mini++;
			    if(mini>=size){mini = 0;}
			}
			else{
			    float d2=(Ai4*Ai4+Ai3*Ai3);
			    if(add && d2*k1*k1>prune_val*prune_val && d2*(1-k1*k1)>prune_val*prune_val){
			    addPoint(Math.round(px[mini]-k1*Ai4),
				     Math.round(py[mini]+k1*Ai3),++mini);
			    }
			    else{
				if(k>0.5){
				    mini++;
				    if(mini==size){mini=0;}
				}
			    }
			    done=true;
			}
		    }
	    }
	    return mini;
    }

	int findClosestPoint(float x,float y, boolean add){
	    float mind = distance2(0,x,y);
	    int mini = 0;
	    for(int i=1;i<size;i++){
		float d = distance2(i,x,y);
		if(d<mind){
		    mini=i;
		    mind=d;
		}
	    }
	    if(add){
	    // add point to snake
		boolean proj = false;
		float xtop = x-px[mini];
		float ytop = y-py[mini];
		int[] intx={0,0};
		int[] inty={0,0};
		float[] dist={0,0};
		for(int j=0;j<2;j++){
		    int k=mini-1+2*j; 		    
		    int l =k<0?size-(Math.abs(k)%size):k%size; //real modulo
		    float xdir = px[l]-px[mini];
		    float ydir = py[l]-py[mini];
		    double norm = Math.sqrt((float)(xdir*xdir+ydir*ydir));
		    xdir /= norm;
		    ydir /= norm;
		    float dp = Math.min((float)norm,Math.max(0,xtop*xdir+ytop*ydir)); 
		    intx[j] = Math.round(px[mini] + dp*xdir);  
		    inty[j] = Math.round(py[mini] + dp*ydir);  
		    dist[j] = (intx[j]-x)*(intx[j]-x)+(inty[j]-y)*(inty[j]-y);
		}
		if(dist[0]<dist[1]){
		    if(distance2(mini,intx[0],inty[0])>=4){
			addPoint(intx[0],inty[0],mini);
		    }
		}
		else{
		    if(distance2(mini,intx[1],inty[1])>=4){
			addPoint(intx[1],inty[1],mini+1);
			mini++;
		    }
		}	
	    }
	    return mini;
	}

	int distance2(int i,int j){//i and j are assumed in the correct range
	    int dx = px[i]-px[j];
	    int dy = py[i]-py[j];
	    return  dx*dx+dy*dy;
	}

	float distance2(int i,float x, float y){//i and j are assumed in the correct range
	    float dx = px[i]-x;
	    float dy = py[i]-y;
	    return  dx*dx+dy*dy;
	}

	double perimeter(int i,int j){//i and j are assumed in the correct range
	    double dx;
	    double dy;
	    double per = distance2(i,j);
	    dx = px[i]-center[0];
	    dy = py[i]-center[1];
	    per += dx*dx+dy*dy;
	    dx = px[j]-center[0];
	    dy = py[j]-center[1];
	    per += dx*dx+dy*dy;
	    return per;
	}
	int[] findTwoOrigins(){
	    updateCenter();
	    int[] or = new int[2];
	    int[] ex = getExtremalPoints();
	    center[0] = 0.5*(center[0]+0.5*(px[ex[0]]+px[ex[1]]));//maybe better?
	    center[1] = 0.5*(center[1]+0.5*(py[ex[0]]+py[ex[1]]));//maybe better?
	    double mind = Double.MAX_VALUE;
	    int mini=0,minj=0;
	    for(int i=ex[0];i<ex[1];i++){
		for(int j=ex[1];j<size+ex[0];j++){
		    int k = j%size;
		    double d = perimeter(i,k);
		    if(d<mind){
			mind = d;
			mini = i;
			minj = k;
		    }
		}
	    }
	    or[0] = Math.min(mini,minj);
	    or[1] = Math.max(mini,minj);
	    return or;
	}
	    // now start gradient descent
	 //    int stop = 0;
	//     Random rand = new Random();
	//     int icount = 10;
	//     int count = icount;
	//     while(count>0){
	// 	stop=1;
	// 	int step = 1+rand.nextInt(5);
	// 	int n = i+step>=ex[1]?ex[1]:i+step;
	// 	int p = i-step<ex[0]?ex[0]:i-step;

	// 	double di = perimeter(p,j);
	// 	if(di<d){
	// 	    d=di;
	// 	    i=p;
	// 	    stop =0;
	// 	}
	// 	else{
	// 	    di = perimeter(n,j);
	// 	    if(di<=d){
	// 		d=di;
	// 		i=n;
	// 		stop = 0;
	// 	    }
	// 	}

	// 	step = 1+rand.nextInt(10);
	// 	n = j+step>=size?j+step-size:j+step;
	// 	if( n>ex[0] && n<ex[1]){n=ex[0];}
	// 	p = j-step<0?size-step+j:j-step;
	//        	if( p>ex[0] && p<ex[1]){n=ex[1];}
	// 	di = perimeter(p,i);
	// 	if(di<=d){
	// 	    d=di;
	// 	    j=p;
	// 	    stop =0;
	// 	}
	// 	else{
	// 	    di = perimeter(n,i);
	// 	    if(di<d){
	// 		j=n;
	// 		d=di;
	// 		stop = 0;
	// 	    }
	// 	}
	// 	if(stop==1){
	// 	    count = count -1;
	// 	}
	// 	else{
	// 	    count = icount;
	// 	}
	//     }
	//     or[0] = Math.min(i,j);
	//     or[1] = Math.max(i,j);
	//     // rois.addRoi(new Line(px[or[0]],py[or[0]],center[0],center[1]));
	//     // rois.addRoi(new Line(center[0],center[1],px[or[1]],py[or[1]]));



	//     return or;
	// }




	void updateCenter(){
	    if(!known_center){
		center[0]=findXCenter();
		center[1]=findYCenter();
		known_center=true;
	    }
	}

	void shrink(double pix){
	    updateCenter();
	    double xc=center[0];
	    double yc=center[1];
	    
   //	    int pixval[4];
	    for(int i=0;i<size;i++){
		double xdir = xc-px[i];
		double ydir = yc-py[i];
		double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		xdir = xdir/norm;
		ydir = ydir/norm;
		double maxj=pix;
		px[i] += Math.round(maxj*xdir);
		py[i] += Math.round(maxj*ydir);
	    }
	}

	int[] getMidShrinkingValue(double pix, ImagePlus im){
	    int[] epix = new int[2];
	    int hwidth = 2;
	    for(int a=0;a<2;a++){
		double xdir = px[mid[1-a]]-px[mid[a]];
		double ydir = py[mid[1-a]]-py[mid[a]];
		double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		xdir = xdir/norm;
		ydir = ydir/norm;
		int i = mid[a];
		int mmax = 0;
		for(int k=i-hwidth;k<=i+hwidth;k++){
		    int l =k<0?size-(Math.abs(k)%size):k%size; //real modulo
		    mmax+= im.getPixel(px[l],py[l])[0];
		}
		int maxj =0;
		
		for(int j=1;j<=pix;j++){
		    int val=0;
		    for(int k=i-hwidth;k<=i+hwidth;k++){
			int l =k<0?size-(Math.abs(k)%size):k%size; //real modulo
			val+= im.getPixel(px[l]+(int)(j*xdir),py[l]+(int)(j*ydir))[0];
		    }
		    if( val>mmax){
			maxj =j;
			mmax = val;
		    }
		}
		epix[a] = maxj;		    
	    }
	    return epix;
	}


	int[] getTipShrinkingValue(double pix, ImagePlus im){
	    double xc = center[0];
	    double yc = center[1];
	    int ex[] = getExtremalPoints();
	    int[] epix = new int[2];	
	    for(int a=0;a<2;a++){
		    int i = ex[a];
		    double xdir = xc-px[i];
		    double ydir = yc-py[i];
		    double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		    xdir = xdir/norm;
		    ydir = ydir/norm;
		    int mmax = im.getPixel(px[i],py[i])[0];
		    int maxj =0;
		    for(int j=(int)(-pix);j<=pix;j++){
			int val = im.getPixel(px[i]+(int)(j*xdir),py[i]+(int)(j*ydir))[0];
			if( val>mmax){
			    maxj =j;
			    mmax = val;
			}
		    }
		    epix[a] = maxj;
		}	    
	    return epix;
	}

	void shrink(double pix, ImagePlus im){
	    shrink(pix,im,1);
	}

	/** 
	 * shrinks the snake to align it on the profile
	 * 
	 * @param pix 
	 * @param im 
	 * @param ref 1: align on pom1 profile, 2:align on cdr2 profile, 3:align on pom1 and cdr2 profiles
	 */
	void shrink(double pix, ImagePlus im, int ref){
	    int[] ex = getExtremalPoints();
	    // rois.addRoi(new Line(px[ex[0]],py[ex[0]],px[ex[1]],py[ex[1]]));
	    int[] epix;
	    updateCenter();
	    double xc = center[0];
	    double yc = center[1];

	    if(ref==1){ 	    // finding the amount of shrinking at extremal points 
		epix = getTipShrinkingValue(pix,im);
	    }
	    else{
		epix = getMidShrinkingValue(pix,im);
	    }
	    int nbp2 =ex[1]-ex[0];
	    int nbp1 = size-ex[1]+ex[0];
	    //	    IJ.log("shrink "+epix[0]+ " "+ epix[1]);
	    for(int i=0;i<size;i++){
		//calculer dir perp. au profile ??
		 double xdir = xc-px[i];
		 double ydir = yc-py[i];
		 double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		 xdir = xdir/norm;
		 ydir = ydir/norm;
		 double fac;
		 if(i<ex[0]){// compter en pixel plutôt qu'en points
		     fac = epix[1]+ ((size-ex[1]+i)/(double)(nbp1))*(epix[0]-epix[1]);    
		 }
		 else{ 
		     if(i<ex[1]){
			 fac = epix[0]+ ((i-ex[0])/(double)(nbp2))*(epix[1]-epix[0]);
		     }
		     else{ //i>=ex[1]
			 fac = epix[1]+ ((i-ex[1])/(double)(nbp1))*(epix[0]-epix[1]);   
		     }
		 }
		 px[i] += (int)(fac*xdir);
		 py[i] += (int)(fac*ydir);
	    }
	    known_center =false;
	}

	void shrink(double pix, ImagePlus im1, ImagePlus im2){
	    int[] ex = getExtremalPoints();
	    updateCenter();
	    double xc = center[0];
	    double yc = center[1];

       	    // finding the amount of shrinking at extremal points 
		int[] epix = getTipShrinkingValue(pix,im1);
	    
       	    // finding the amount of shrinking at mid-cells 
		int[] midpix = getMidShrinkingValue(pix,im2);
	    
	    boolean mfirst = mid[0]<ex[0];
	    int[] pos = new int[4];
	    int[] val = new int[4];
	    if(mfirst){
		pos[0] = mid[0];pos[1] = ex[0];pos[2] = mid[1];pos[3] = ex[1];
		val[0] = midpix[0];val[1] = epix[0];val[2] = midpix[1];val[3] = epix[1];
	    }
	    else{
		pos[0] = ex[0];pos[1] = mid[0];pos[2] = ex[1];pos[3] = mid[1];
		val[0] = epix[0];val[1] = midpix[0];val[2] = epix[1];val[3] = midpix[1];
	    }
	    IJ.log("shrink "+epix[0]+ " "+ epix[1]+" "+midpix[0]+ " "+ midpix[1]);
	    for(int i=0;i<size;i++){
		//calculer dir perp. au profile ??
		 double xdir = xc-px[i];
		 double ydir = yc-py[i];
		 double norm = Math.sqrt(xdir*xdir+ydir*ydir);
		 xdir = xdir/norm;
		 ydir = ydir/norm;
		 double fac;
		 if(i<=pos[0]){
		     fac = val[3]+ ((size-pos[3]+i)/(double)(size-pos[3]+pos[0]))*(val[0]-val[3]);    
		 }
		 else{
		     if(i<=pos[1]){
			 fac = val[0]+ ((i-pos[0])/(double)(pos[1]-pos[0]))*(val[1]-val[0]);    
		     }
		     else{
			 if(i<=pos[2]){
			     fac = val[1]+ ((i-pos[1])/(double)(pos[2]-pos[1]))*(val[2]-val[1]);    
			 }
			 else{
			     if(i<=pos[3]){
				 fac = val[2]+ ((i-pos[2])/(double)(pos[3]-pos[2]))*(val[3]-val[2]);
			     }
			     else{
				 fac = val[3]+ ((i-pos[3])/(double)(size-pos[3]+pos[0]))*(val[0]-val[3]);
			     }
			 }
		     }
		     
		 }
		 px[i] += (int)(fac*xdir);
		 py[i] += (int)(fac*ydir);
	    }
	    known_center =false;
	}
	double findXCenter(){ 
	    double s=0;
	    for(int i=0;i<size;i++){s+=px[i];}
	    return s/size; //assumes homoeneous point distribution
	}

	double findYCenter(){
	    double s=0;
	    for(int i=0;i<size;i++){s+=py[i];}
	    return s/size; //assumes homoeneous point distribution
	}

	
	double getLength(){
	    int[] ex= getExtremalPoints();
	    int dxi = px[ex[0]]-px[ex[1]];
	    int dyi = py[ex[0]]-py[ex[1]];
	    double d= dxi*dxi+dyi*dyi;
	    return Math.sqrt(d);
	}

	void setMidPoints(int i,int j){
	    mid[0]=i;
	    mid[1]=j;
	    known_mid =true;
	}

	int[] getMidPoints(){
	    if(known_mid){
		return mid;
	    }
	    else{
		//we could actually infer it
		return null;
	    }
	}

	void setExtremalPoints(int i,int j){
	    extr[0]=i;
	    extr[1]=j;
	    center[0] = 0.5*(px[extr[0]]+px[extr[1]]);
	    center[1] = 0.5*(py[extr[0]]+py[extr[1]]);
	    known_extr=true;
	    known_center=true;
	}

	int[] getExtremalPoints(){
	    if(known_extr){
		return extr;
	    }

	    // brute force
	    float maxd=0;
	    int maxi=0;
	    int maxj=0;
	    for(int i=0;i<size;i++){
		for(int j=0;j<size;j++){
		    float d = distance2(i,j);
		    if(d>maxd){
			maxd=d;
			maxi=i;
			maxj=j;
		    }
		}
	    }
	    extr[0] = Math.min(maxi,maxj);
	    extr[1] = Math.max(maxi,maxj);
	    known_extr=true;
	    return extr;
	}
    }
    
    //    public enum State{INIT,SELECT,VALID};
   
    ImagePlus img;
    ImagePlus profImg;
    ImagePlus cdr2Img;
    ImageCanvas canvas;
    ImageStack stack;
    EllipseFitter fitter;
    //  Wand wand;
    RoiManager rois;
    ImageProcessor ip;
    Snake snake;
    Hashtable<Integer,Snake> allsnakes;
    Edges edges;
    Image edgim;
    String filename;
    String basename;
    String dir;
    int start;
    int towrite;
    PrintWriter out;
    PrintWriter posout;

    boolean showback;
    boolean threeChannels;

    //  State state;
    final int NPOINTS = 80;

    public int setup(String arg, ImagePlus img) {
		this.img = img;	
		IJ.register(CellophaneManualNoDext_.class);
		rois = RoiManager.getInstance();
		edges = new Edges();

		if(rois ==null){
		    rois =new RoiManager();
		}
		fitter = new EllipseFitter();
		stack = img.getImageStack();
		filename = img.getOriginalFileInfo().fileName;
		dir = img.getOriginalFileInfo().directory;
		start =0;
		towrite = 1;
		showback = false;
		return DOES_ALL;
    }
    public void initFiles(String imagefname){
	int n= imagefname.lastIndexOf(".");
	if(n>=0){
	    basename = imagefname.substring(0,n);
	}
	else{
	    basename = imagefname;
	}
	FileWriter fw;
	basename = dir+basename;
	try{
	    fw= new FileWriter(basename+"_profiles.csv");
	    out = new PrintWriter(new BufferedWriter (fw));
	}catch(Exception e){
	    IJ.log("can't open file: "+e);
	}
	FileWriter pfw;
	try{
	    pfw= new FileWriter(basename+"_position.csv");
	    posout = new PrintWriter(new BufferedWriter (pfw));
	}catch(Exception e){
	    IJ.log("can't open file: "+e);
	}	
    }

    public float fpDistance2(FloatPolygon p1, int i, FloatPolygon p2, int j){
	float dx = p1.xpoints[i]-p2.xpoints[j];
	float dy = p1.ypoints[i]-p2.ypoints[j];	
	return dx*dx+dy*dy;
    }

    public int polClosestPoint(FloatPolygon p, int x, int y){
	float mind = Float.MAX_VALUE;
	float d2;
	int mini = -1;
	for(int i=0;i<p.npoints;i++){
	    float dx = p.xpoints[i]-x;
	    float dy = p.ypoints[i]-y;	
	    d2= dx*dx+dy*dy;
	    if(d2<mind){
		mind=d2;
		mini=i;
	    }
	}
	return mini;
    }


    public void run(ImageProcessor ip) {
	//	IJ.log("file info" + IJ.getImage().getOriginalFileInfo().toString());
	//	initFiles(IJ.getImage().getTitle());
	String stackname =IJ.getImage().getTitle();
	this.ip = ip.convertToFloat();
	//	threeChannels = (img.getNSlices()%2)==1;
    	threeChannels = false; // this is a pom1tomato-cdr2GFP version
	ImagePlus profimg = img.duplicate();//could probably be more efficient
	ImagePlus cdr2img=img.duplicate();//could probably be more efficientnull; 


	for (int s=2;s<=cdr2img.getNSlices();s++){//removes every other images
	    cdr2img.getStack().deleteSlice(s);
	}
	// ImageStack stack = profimg.getStack();
	// stack.deleteSlice(1);
	// profimg.setStack(null, stack); 	
	for (int s=1;s<profimg.getNSlices();s++){//removes every other images
	    profimg.getStack().deleteSlice(s);
	}


	
	//	IJ.log("nslices "+profimg.getNSlices()+ "  "+ deximg.getNSlices()+ "  "+ cdr2img.getNSlices());

	ZProjector profproj = new ZProjector(profimg);
	profproj.setMethod(ZProjector.AVG_METHOD);
	profproj.setStartSlice(1);
	profproj.setStopSlice(profimg.getNSlices());
	profproj.doProjection();
	profImg = profproj.getProjection();

	ZProjector cdr2proj = new ZProjector(cdr2img);
	cdr2proj.setMethod(ZProjector.AVG_METHOD);
	cdr2proj.setStartSlice(1);
	cdr2proj.setStopSlice(cdr2img.getNSlices());
	cdr2proj.doProjection();
	cdr2Img = cdr2proj.getProjection();
	    
	 cdr2Img.show();
	 profImg.show();

	 // ImagePlus timp = new ImagePlus("test2",profimg.getStack());
	 // timp.show();
       	// cdr2img.show();
	// profimg.show();

	ImageWindow win = profImg.getWindow();
	canvas = win.getCanvas();
	canvas.addKeyListener(this) ; 
	// cdr2Img.getWindow().getCanvas().addKeyListener(this);
	// profImg.getWindow().getCanvas().addKeyListener(this);
	start=0;
	IJ.log("Draw profiles");
	IJ.log("Press 'Ctl-t' to validate and <space>  when you are done (and no ROI is selected)");
	while(start==0){
	    try{
		Thread.sleep(500);
	    }
	    catch(Exception e){		IJ.log("exception"+e);}
	}
	
	//initializing snakes
	IJ.log("Done");
	Hashtable ohash = rois.getROIs();
	Hashtable hash = (Hashtable) ohash.clone();

	Enumeration e = hash.elements();

	int snakecnt = 0;
	allsnakes = new Hashtable<Integer,Snake>();
	while(e.hasMoreElements()){
	    snakecnt++;
	    snake = new Snake(3);
	    Roi r = (Roi) e.nextElement();
	    // r.setName("p"+snakecnt);
	    // r.setStrokeWidth(snake.STROKE_WIDTH);
	    snake.getPointsFromRoi(r,NPOINTS);
	    Roi r2 = snake.asRoi();
	    r2.setName("p"+snakecnt);
	    rois.addRoi(r2);
	    int[] ext = snake.getExtremalPoints();
	    Roi bigax = new Line(snake.getX(ext[0]),snake.getY(ext[0]),
				 snake.getX(ext[1]),snake.getY(ext[1]));
	    bigax.setName("p"+snakecnt+"_b");
	    rois.addRoi(bigax);
	    int[] or = snake.findTwoOrigins();
	    Roi smallax = new Line(snake.getX(or[0]),snake.getY(or[0]),
				   snake.getX(or[1]),snake.getY(or[1]));
	    smallax.setName("p"+snakecnt+"_s");
	    rois.addRoi(smallax);
	    allsnakes.put(snakecnt,snake);
	}
	IJ.log("we have "+snakecnt+" profiles");


	start =1;

	IJ.log("Please correct the cell axes if necessary by selecting them and updating them with the ROI Manager");

	IJ.log("Press <space> to confim");
	while(start==1){
	    try{
	 	Thread.sleep(500);
	    }
	    catch(Exception me){IJ.log("exception"+me);}
	}
	towrite = 1;
	while(towrite==1){

	initFiles(stackname);
	// // validate
	int nbsnakes = snakecnt;
	int nbRoiM = rois.getCount();
	IJ.log("Writing profiles");
	int nbprof = 0;
      for(int ss=1;ss<=nbsnakes;ss++){
	    //	    IJ.log("looking for snake "+ss);
	    int ind=-1;
	    int inds = -1; //index of small axis
	    int indb = -1; //index of big axis
	    String name = "p"+ss;
	    String nameb = "p"+ss+"_b";
	    String names = "p"+ss+"_s";
	    for(int i=0;i<nbRoiM;i++){
		if(name.equals(rois.getName(i))){
		    ind =i;
		    continue;
		}
		if(nameb.equals(rois.getName(i))){
		    indb =i;
		    continue;
		}
		if(names.equals(rois.getName(i))){
		    inds =i;
		    continue;
		}
	    }
	    if(ind==-1){ //snake not found in roi manager
		IJ.log("profile "+ss +" not found ");	
		continue;
	    }
	    // extract points from roi manager
	    snake = allsnakes.get(ss);
	    snake.getPointsFromRoi(rois.getRoisAsArray()[ind],-1);
	    FloatPolygon fp=(rois.getRoisAsArray()[inds]).getFloatPolygon();
	    int[] or=new int[2];
	    // or[0] = snake.findClosestPoint(fp.xpoints[0],fp.ypoints[0],true);
	    // or[1] = snake.findClosestPoint(fp.xpoints[1],fp.ypoints[1],true);
	    or[0] = snake.findClosestPoint(fp.xpoints[0],fp.ypoints[0],fp.xpoints[1],
					   fp.ypoints[1],true);
	    or[1] = snake.findClosestPoint(fp.xpoints[1],fp.ypoints[1],fp.xpoints[0],
					   fp.ypoints[0],true);
	    snake.setMidPoints(or[0],or[1]);
	    fp = (rois.getRoisAsArray()[indb]).getFloatPolygon();
	    int[] extr=new int[2];
	    // extr[0] = snake.findClosestPoint(fp.xpoints[0],fp.ypoints[0],true);
	    // extr[1] = snake.findClosestPoint(fp.xpoints[1],fp.ypoints[1],true);
	    extr[0] = snake.findClosestPoint(fp.xpoints[0],fp.ypoints[0],fp.xpoints[1],
					     fp.ypoints[1],true);
	    extr[1] = snake.findClosestPoint(fp.xpoints[1],fp.ypoints[1],fp.xpoints[0],
					     fp.ypoints[0],true);
	    snake.setExtremalPoints(extr[0],extr[1]);
	    //    IJ.log("extr: "+ extr[0]+ " " + extr[1]+ " " + or[0]+ " " + or[1]);
	    double si = snake.getLength(); 
	    double[] backgr = snake.getExtBackground(10,profImg,showback);
	    double[] cybak = snake.getExtBackground(-10,profImg,false);
	    //get cdr2 background
	    double[] cdr2backgr= snake.getExtBackground(10,cdr2Img,showback);
	    double[] cdr2cybak = snake.getExtBackground(-10,cdr2Img,false);
	    PolygonRoi[] sroi = new PolygonRoi[4];
	    sroi[0] = snake.asRoi(or[0],true);
	    rois.addRoi(sroi[0]);
	    sroi[1] = snake.asRoi(or[0],false);
	    sroi[2] = snake.asRoi(or[1],true);
	    sroi[3] = snake.asRoi(or[1],false);
	    for(int j=0;j<4;j++){
		profImg.setRoi(sroi[j]);
		ProfilePlot prof = new ProfilePlot(profImg,true);
		double[] profval = prof.getProfile();
		cdr2Img.setRoi(sroi[j]);
		ProfilePlot cdr2prof = new ProfilePlot(cdr2Img,true);
		double[] cdr2profval = cdr2prof.getProfile();
		FloatPolygon p = sroi[j].getFloatPolygon();
		int mink0 = polClosestPoint(p,snake.getX(extr[0]),snake.getY(extr[0]));
		int mink1 = polClosestPoint(p,snake.getX(extr[1]),snake.getY(extr[1]));

		out.print(filename+";"+ss+";"+j+";"+si+";"+profval.length
			  +";"+ "1;"+backgr[0] +";"+ backgr[1] +";"+ cybak[0] +";"+
			  cybak[1] +";"+ mink0 +";"+ mink1);
		for(int k=0;k<profval.length;k++){//printing values
		    out.print(";"+(profval[k]-backgr[0]));
		}
		out.println();

		out.print(filename+";"+ss+";"+j+";"+si+";"+cdr2profval.length
			  +";"+ "2;"+cdr2backgr[0] +";"+ cdr2backgr[1] +";"+ cdr2cybak[0] +";"+
			  cdr2cybak[1] +";"+ mink0 +";"+ mink1);
		try{
		for(int k=0;k<cdr2profval.length;k++){//printing values
		    double relval = cdr2profval[k]-cdr2backgr[0];
		    out.print(";"+relval);
		}
		out.println();
		}
		catch(Exception pe){
		    IJ.log("cannot write full profile #"+ ss);
		    IJ.log("exception: "+pe);
		}
		posout.print(ss+";"+j+";"+si+";"+p.npoints);
		for(int k=0;k<p.npoints;k++){//printing values
		    posout.print(";"+p.xpoints[k]);
		}
		posout.println();
		posout.print(";"+";"+";");
		for(int k=0;k<p.npoints;k++){//printing values
		    posout.print(";"+p.ypoints[k]);
		}
		posout.println();
	    }
	    // writing roi
	    try{
		RoiEncoder re = new RoiEncoder(basename+"_"+ss+".roi");
		re.write(sroi[0]);
	    }catch(IOException ioe){
		IJ.log("cannot save roi in "+basename+"_"+ss);
		IJ.error("ROI Writer", ioe.getMessage());
	    }
	    nbprof = nbprof+1;
	    IJ.log("stats "+si+" "+backgr[0]+ " +- "+ backgr[1]);
	    
	}
      towrite=0;
      IJ.log(" ");
      IJ.log("wrote "+ nbprof + " profiles. Click <space> to confirm or 't' if you expected more and want to try again");
	start = 0;
	while(start==0){
	    try{
	 	Thread.sleep(500);
	    }
	    catch(Exception me){IJ.log("exception"+me);}
	}
	out.close();
	posout.close();	
	}
	IJ.log("done");
	canvas.removeKeyListener(this);
    }

    public void mouseClicked(MouseEvent e) {
    }
   
    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {}
    public void mouseReleased(MouseEvent e) {}

    public void keyTyped(KeyEvent e){
	if(e.getKeyChar()==' '){
	    start=1-start;
	}
	if(e.getKeyChar()=='b'){
	    IJ.log("showing background extraction");
	    showback = true;
	}
	if(e.getKeyChar()=='v'){
	    IJ.run("Add to Manager");
	    //    r.update(true,false);
	}

	if(e.getKeyChar()=='t'){
	    towrite=1;
	    start=1-start;
	     
	}
    } 
    public void keyPressed(KeyEvent e){}
    public void keyReleased(KeyEvent e){}
	
}



 

